package sun.rmi.transport;

import java.util.HashMap;
import sun.security.action.GetLongAction;
import java.security.PrivilegedAction;
import java.security.AccessController;
import sun.rmi.runtime.NewThreadAction;
import java.rmi.dgc.VMID;
import java.rmi.server.ObjID;
import java.rmi.server.ExportException;
import sun.rmi.runtime.Log;
import java.rmi.NoSuchObjectException;
import java.rmi.Remote;
import sun.misc.GC;
import java.lang.ref.ReferenceQueue;
import java.util.Map;

public final class ObjectTable
{
    private static final long gcInterval;
    private static final Object tableLock;
    private static final Map<ObjectEndpoint, Target> objTable;
    private static final Map<WeakRef, Target> implTable;
    private static final Object keepAliveLock;
    private static int keepAliveCount;
    private static Thread reaper;
    static final ReferenceQueue<Object> reapQueue;
    private static GC.LatencyRequest gcLatencyRequest;
    
    private ObjectTable() {
    }
    
    static Target getTarget(ObjectEndpoint objectEndpoint) {
        synchronized (ObjectTable.tableLock) {
            return (Target)ObjectTable.objTable.get(objectEndpoint);
        }
    }
    
    public static Target getTarget(Remote remote) {
        synchronized (ObjectTable.tableLock) {
            return (Target)ObjectTable.implTable.get(new WeakRef(remote));
        }
    }
    
    public static Remote getStub(Remote remote) throws NoSuchObjectException {
        Target target = getTarget(remote);
        if (target == null) {
            throw new NoSuchObjectException("object not exported");
        }
        return target.getStub();
    }
    
    public static boolean unexportObject(Remote remote, boolean b) throws NoSuchObjectException {
        synchronized (ObjectTable.tableLock) {
            Target target = getTarget(remote);
            if (target == null) {
                throw new NoSuchObjectException("object not exported");
            }
            if (target.unexport(b)) {
                removeTarget(target);
                return true;
            }
            return false;
        }
    }
    
    static void putTarget(Target target) throws ExportException {
        ObjectEndpoint objectEndpoint = target.getObjectEndpoint();
        WeakRef weakImpl = target.getWeakImpl();
        if (DGCImpl.dgcLog.isLoggable(Log.VERBOSE)) {
            DGCImpl.dgcLog.log(Log.VERBOSE, "add object " + objectEndpoint);
        }
        synchronized (ObjectTable.tableLock) {
            if (target.getImpl() != null) {
                if (ObjectTable.objTable.containsKey(objectEndpoint)) {
                    throw new ExportException("internal error: ObjID already in use");
                }
                if (ObjectTable.implTable.containsKey(weakImpl)) {
                    throw new ExportException("object already exported");
                }
                ObjectTable.objTable.put(objectEndpoint, target);
                ObjectTable.implTable.put(weakImpl, target);
                if (!target.isPermanent()) {
                    incrementKeepAliveCount();
                }
            }
        }
    }
    
    private static void removeTarget(Target target) {
        ObjectEndpoint objectEndpoint = target.getObjectEndpoint();
        WeakRef weakImpl = target.getWeakImpl();
        if (DGCImpl.dgcLog.isLoggable(Log.VERBOSE)) {
            DGCImpl.dgcLog.log(Log.VERBOSE, "remove object " + objectEndpoint);
        }
        ObjectTable.objTable.remove(objectEndpoint);
        ObjectTable.implTable.remove(weakImpl);
        target.markRemoved();
    }
    
    static void referenced(ObjID objID, long n, VMID vmid) {
        synchronized (ObjectTable.tableLock) {
            Target target = (Target)ObjectTable.objTable.get(new ObjectEndpoint(objID, Transport.currentTransport()));
            if (target != null) {
                target.referenced(n, vmid);
            }
        }
    }
    
    static void unreferenced(ObjID objID, long n, VMID vmid, boolean b) {
        synchronized (ObjectTable.tableLock) {
            Target target = (Target)ObjectTable.objTable.get(new ObjectEndpoint(objID, Transport.currentTransport()));
            if (target != null) {
                target.unreferenced(n, vmid, b);
            }
        }
    }
    
    static void incrementKeepAliveCount() {
        synchronized (ObjectTable.keepAliveLock) {
            ++ObjectTable.keepAliveCount;
            if (ObjectTable.reaper == null) {
                (ObjectTable.reaper = (Thread)AccessController.doPrivileged(new NewThreadAction(new ObjectTable.Reaper(), "Reaper", false))).start();
            }
            if (ObjectTable.gcLatencyRequest == null) {
                ObjectTable.gcLatencyRequest = GC.requestLatency(ObjectTable.gcInterval);
            }
        }
    }
    
    static void decrementKeepAliveCount() {
        synchronized (ObjectTable.keepAliveLock) {
            --ObjectTable.keepAliveCount;
            if (ObjectTable.keepAliveCount == 0) {
                if (ObjectTable.reaper == null) {
                    throw new AssertionError();
                }
                AccessController.doPrivileged(new PrivilegedAction<Void>() {
                    @Override
                    public Void run() {
                        ObjectTable.reaper.interrupt();
                        return null;
                    }
                });
                ObjectTable.reaper = null;
                ObjectTable.gcLatencyRequest.cancel();
                ObjectTable.gcLatencyRequest = null;
            }
        }
    }
    
    static void access$400(Target target) {
        removeTarget(target);
    }
    
    static {
        gcInterval = (long)AccessController.doPrivileged(new GetLongAction("sun.rmi.dgc.server.gcInterval", 3600000L));
        tableLock = new Object();
        objTable = new HashMap();
        implTable = new HashMap();
        keepAliveLock = new Object();
        ObjectTable.keepAliveCount = 0;
        ObjectTable.reaper = null;
        reapQueue = new ReferenceQueue();
        ObjectTable.gcLatencyRequest = null;
    }
    
    private static class Reaper implements Runnable
    {
        @Override
        public void run() {
            try {
                do {
                    WeakRef weakRef = (WeakRef)ObjectTable.reapQueue.remove();
                    synchronized (ObjectTable.tableLock) {
                        Target target = (Target)ObjectTable.implTable.get(weakRef);
                        if (target != null) {
                            if (!target.isEmpty()) {
                                throw new Error("object with known references collected");
                            }
                            if (target.isPermanent()) {
                                throw new Error("permanent object collected");
                            }
                            ObjectTable.access$400(target);
                        }
                    }
                } while (!Thread.interrupted());
            }
            catch (InterruptedException ex) {}
        }
    }
}
